#########################################################################################
#########################################################################################
## 
## Coercion, Information, and the Success of Sanction Threats
## - Taehee Whang, Elena V. McLean, and Douglas Kuberski
## 
## 1. Change the working directory.
## 2. Make sure to save the lib_cPBE folder in the working directory.
##
## Taehee Whang
##  2012/05/24 AJPS Supporting Information  
##  - The model part is run using a separate library based on fortran.
##  - The main part of the fortran code is also included at the end of this R code.
##  - For any questions, please contact the authors.
##
#########################################################################################
#########################################################################################


# Change the working directory
rm(list = ls())
library(foreign)
setwd("c:/Users/taeheewhang/Dropbox/McLean_prj3/AJPS final/Replication package")  # working directory here
libname="lib_cPBE/Release/lib_cPBE.dll" # windows
#libname="lib_cPBE/Release/lib_cPBE.so" # unix (for linux cluster)

LinuxCluster = 0
# if 1, snow library is used.
# if 0, "fake" snow library is used so that snow function works without real cluster.

##########
## Data ##
##########

g_datf = read.dta("data_replication.dta") # data here
g_datf['const']=1  # add a constant column
g_datf['zero']=0   # add a null column
attach(g_datf)

#####################################################################
## Model                                                           ##
## - Equilibrium probability calculations run in lib_cPBE (fortran)## 
## - We have non-vectorized version for R but it is very slow      ##
## - We provide the main part of lib_cPBE at the end of the R code ##
#####################################################################

# calculate equilibrium probability - a vectorized version.
get_pr_vec <-function(psets_t,tol=.Machine$double.eps^0.5){
  psets = t(psets_t)
  n = nrow(psets_t)
  pfrc = .C("solve_eq_cpbe__",
      as.integer(n),
          as.double(psets),
      as.double(tol),
          pfrc = double(n*4))$pfrc
  pfrc = matrix(pfrc,nrow=4,ncol=n)
  pfrc = t(pfrc)
  colnames(pfrc) <- c('pf','pr','pc','pf_prior')
  return(pfrc)
}

# calculate equilibrium probability - pset is assumed to be an array.
get_pr <- function(pset,tol=.Machine$double.eps^0.5){
    return(get_pr_vec(t(pset))[1,])
}

################
## Estimation ##
################

estimate <-function(pset0, method="Nelder-Mead") {

    optim_result = optim(pset0, likelihood_model_snow,
                method = method,   # "Nelder-Mead", "BFGS", "CG", "L-BFGS-B", "SANN"
                control = list(trace=0,maxit=10^5) # ,reltol=1e-10)
                )
print(optim_result)
#   if (optim_result$convergence!=0) {
#       print(paste("ERR : ",
#               optim_result$convergence,
#               " & ",
#               optim_result$message
#               )
#           )
#   }

    pset_hat = optim_result$par
    names(pset_hat) <- g_reg_pset_names
    return(pset_hat)
}

# likelihood is actually log-likelihood
likelihood_model <- function(reg_pset,dataf)
  sum(likelihood_individual(reg_pset,dataf))

# likelihood of each observation
likelihood_individual <-function(reg_pset,dataf){
    psets = make_pset(reg_pset,dataf)  # calculate expected values for each node.
    lik = matrix(0, nrow=nrow(dataf), ncol=1)
    lik_penalty = matrix(g_penalty , nrow=nrow(dataf), ncol=1)

    pfrc = get_pr_vec(psets) # equilibrium probabilities
    if (pfrc[1,1]< 0) return(lik_penalty)

    psq = 1 - pfrc[,3]
    pcd = pfrc[,3] * (1-pfrc[,2])
    pbd = pfrc[,3] * pfrc[,2] * (1-pfrc[,1])
    psf = pfrc[,3] * pfrc[,2] * pfrc[,1]

        z=1:nrow(dataf)
        lik[dataf$sq * z] = log(psq[dataf$sq * z])
        lik[dataf$acq * z] = log(pcd[dataf$acq * z])
        lik[dataf$bd * z] = log(pbd[dataf$bd * z])
        lik[dataf$sanctions * z] = log(psf[dataf$sanctions * z])
    # lik = log(psq)*dataf$sq + log(pcd)*dataf$acq + log(pbd)*dataf$bd + log(psf)*dataf$sanctions
    # log-likelihood
    if (all(is.finite(lik)))
        return(-lik)
    else
        return(lik_penalty)
}

likelihood_at_each_node <- function(p) {
      ll = likelihood_model(p,g_datf_sub) #g_datf_sub should be assigned at each node before using this function
      return(sum(unlist(ll)))
    }

# likelihood using snow
likelihood_model_snow <- function(reg_pset){
  lik_list = clusterCall(cl,likelihood_at_each_node,reg_pset)
  if (all(lik_list<g_penalty*nrow(g_datf)/(length(cl)+1))) {
    return( sum(unlist(lik_list)))
  } else {
    return( g_penalty * nrow(g_datf) )
  }
}

# calculate likelihood for each observation, using snow
likelihood_individual_snow <- function(reg_pset){
  lik_list = clusterCall(cl,
    function(p) likelihood_individual(p,g_datf_sub),reg_pset)
  return( unlist(lik_list))
}

# make model parameters, based on regressors and given coef.
# using global variables, g_reg_names and g_pset_names
make_pset <- function(reg_pset,dataf){
    psets = matrix(0,nrow = nrow(dataf), ncol = length(g_pset_names) )
    colnames(psets) <- g_pset_names
    for (i in 1:length(g_pset_names)){
        for (j in 1: (length(g_reg_names)/2)  ){
            if (g_pset_names[i]==g_reg_names[j*2-1]) {
                psets[,i] = psets[,i] + data.matrix(dataf[g_reg_names[j*2]]) * reg_pset[j]
                        }
        }
    }
    # do not estimate sVWB and sVVB
    psets[,'sVVB']=0
    psets[,'sVWB']=0

    return(psets)
}

# variance estimates, based on BHHH(outer product method)
get_asymptiotic_var <- function(lnlik,x,n_obs, dif = (.Machine$double.eps)^0.5){ 
# lnlik : a function of a parameter that returns a column vector of log likelihood for each observation.
# x : MLE parameter
  G = matrix(0,nrow = n_obs, ncol = length(x) )
  for (i in 1:length(x)){
    x0 = x
    x0[i] = x0[i] - dif/2
    x1 = x
    x1[i] = x1[i] + dif/2
    y0 = lnlik(x0)
    y1 = lnlik(x1)
    G[,i] = (y1 - y0)/ dif
  }
  BHHH = solve(t(G) %*% G )
  return(BHHH)
}

# make each parameters name.
get_reg_pset_names <- function(reg_names){
    reg_pset_names = rep('',times=length(reg_names)/2)
    for (i in 1:(length(reg_names)/2))
        reg_pset_names[i] = paste(reg_names[i*2-1],reg_names[i*2],sep="_")
    return(reg_pset_names)
}

# belief updating for challenge case
belief_updating <- function(reg_pset,dataf){
    psets = make_pset(reg_pset,dataf)  # calculate expected values for each node.
    pfrc = get_pr_vec(psets) # equilibrium probabilities
    if (pfrc[1,1]< 0) return("ERR")
    
    z=1:nrow(dataf)
    bu = pfrc[,1] - pfrc[,4]

    bu_challange = bu [ (1-dataf$sq) * z ]
    return(bu_challange)
}
  
##########################
#     INITIALIZATION     #
##########################

ncl = 5 # number of clusters. When using fake version, it will be changed to 1.
if (LinuxCluster==1)
    {
    library(snow)
    } else  # snow function is fake now.
    {
        ncl = 1  # there is no actual cluster.
        makeCluster <- function(ncl){
            return(list(1))   # only one node
        }
        
        clusterCall <- function(cl,fun,...){
            temp = do.call(fun,list(...))
            return(list(temp))
        }

        clusterEvalQ <- function(cl,fun){
            temp = eval(fun)
            return(list(temp))
        }

        clusterSetupSPRNG <- function(cl, seed){
            set.seed(seed)
        }

        clusterApply <- function(cl,X, FUN, ...){
            lapply(X, FUN, ...)
        }

        clusterSplit <- function(cl,x) { list(x) }

        clusterExport <- function(cl,li){ }
    }

cl = makeCluster(ncl)

library(foreign)
library(nlme)
temp = clusterEvalQ(cl,library(MASS))

di<-dyn.load(libname)
temp = clusterCall(cl,"dyn.load",libname)

# stopCluster(cl)
.C("init_cpbe_dll__")    # initialize dll
temp = clusterCall(cl,".C","init_cpbe_dll__")


##################
##     run      ##
##################

g_pset = 0   # for debugging

g_penalty = 100 # if 0 probability is assigned, this value is given as a penalty log-likelihood
g_pset_names = c    ('mSA','mVA','maA','mWA','sVVA','sVaA','sVWA','saWA','sWWA','mCB','mVB','mWB','sVVB','sVWB') # used in make_pset. Do not change the order.
pset1 =     c     (    0,    1, -1/2,   -1,     0,     0,     0,     0,     1,    0,    1,   -1,     0,     0)
names(pset1) <-g_pset_names

## basic info ##
true_pr1 = get_pr(pset1)
print( true_pr1 )
cat("pf diff.=", true_pr1['pf']-true_pr1['pf_prior'],"\n")

# regressor names - this consists of pairs of (nodename,regressor).
# nodename - one of g_pset_names
# regressor - one of column names in data.frame

g_reg_names =   c('mSA','lndxmo1',      
            'mSA','polity21',   
            'mSA','contig', 
            'mSA','alliance2',
            
            'mVA','const',              
            'mVA','anticipatedsendercosts', 

            'maA','const',      
            'maA','polity21',   
            
            'mWA','const',      
            'mWA','lndxmo1',    
            'mWA','polity21',   
            'mWA','lncaprat',   

            'mCB','const',              
            'mCB','lndxmo2',            
            'mCB','anticipatedtargetcosts',
            'mCB','contig',         
            'mCB','alliance2',          
        
            'mWB','const',          
            'mWB','polity22',       
            'mWB','lncaprat',       

            'sVVA','const',
            'sVaA','const',
            'sVWA','const',
            'saWA','const',
            'sWWA','const')

g_reg_pset_names = get_reg_pset_names(g_reg_names) # for display later

reg_pset1 = c( -2.000000e-01, -3.914750e-04,  8.504790e-04, -1.086818e-04,  6.182462e-01,  8.890048e-05, -1.856716e-05, -5.627585e-05, -8.478712e-05, -2.000000e-01, 
-4.847100e-04,  2.789460e-04, 5.154539e-05,  3.000000e-01,  8.698855e-05,  
0.000000e+00,  0.000000e+00,  7.928999e-05, -4.847100e-04,  2.789460e-04,  
1.000000e-02,  0.000000e+00, 0.000000e+00,  0.000000e+00, 1.000000e-01)

temp=clusterExport(cl,ls()) # prepare for snow
g_sub_list = clusterSplit(cl,1:nrow(g_datf)) # list of sequences, for each node.
g_datf_sub_list = lapply(g_sub_list, function(i) g_datf[i,]) # list of sub_data. Each node will process one sub_data.
temp=clusterApply(cl,g_datf_sub_list,
  function(x) assign("g_datf_sub",x,envir = .GlobalEnv)) # Each node gets its own sub_data as a global variable.
print(unlist(clusterCall(cl,function() nrow(g_datf_sub)))) # number of observations that each node will process.

reg_pset0 = list()
lik0 = list()
reg_pset0[[1]] = estimate(reg_pset1)
reg_pset0[[2]] = estimate(reg_pset0[[1]],"BFGS")
lik0[[1]] = likelihood_model(reg_pset0[[1]],g_datf)
lik0[[2]] = likelihood_model(reg_pset0[[2]],g_datf)

for (i in 1:5){
    i=i*2+1
    reg_pset0[[i]] = estimate(reg_pset0[[i-1]])
    reg_pset0[[i+1]] = estimate(reg_pset0[[i]],"BFGS")
    lik0[[i]] = likelihood_model(reg_pset0[[i]],g_datf)
    lik0[[i+1]] = likelihood_model(reg_pset0[[i+1]],g_datf)
}

make_pfrc <- function(reg_pset,dataf){
    psets = make_pset(reg_pset,dataf)  # calculate expected values for each node.
    pfrc = get_pr_vec(psets) # equilibrium probabilities
    return(pfrc)
}

pfrc = make_pfrc(reg_pset0[[12]],g_datf)
log_likelihood = likelihood_model(reg_pset0[[12]],g_datf)
reg_pset_KMW = reg_pset0[[12]]
aV=get_asymptiotic_var(likelihood_individual_snow,reg_pset_KMW,nrow(g_datf))
reg_pset_KMW_sd = diag(aV) ^ 0.5
est_summary = cbind(as.matrix(reg_pset_KMW),reg_pset_KMW_sd, as.matrix(reg_pset_KMW/reg_pset_KMW_sd) )

est_summary 
log_likelihood

save(reg_pset_KMW,aV,est_summary,pfrc,reg_pset1,log_likelihood, file="est_KMW_ajps_replication.Rsave")

####################################################################################################



#
#################################################################################
### lib_cPBE - main part of the fortran code                                   ##
### - Both the R code above and lib_cPBE folder will be available on the web   ##
#################################################################################
#
#Module global_variables
#    integer mSA,mVA,maA,mWA,sVVA,sVaA,sVWA,saWA,sWWA,mCB,mVB,mWB,sVVB,sVWB
#    real g_pset(14)
#end Module
#
#subroutine get_cdfbvn(x,y,rhom,p)
#USE Bivariate_Trivariate_Normal
#Implicit none
#real  x,y,rhom
#real, INTENT(OUT):: p
#    p = bvn(-x,-y,rhom)
#End subroutine
#
#subroutine solve_eq_cPBE(n,psets,tol,pfrc)
#USE global_variables
#Implicit none
#integer n
#real psets(14,n), tol
#real, INTENT(OUT):: pfrc(4,n)
#integer i
#real pf, p3(3), prior
#real, external:: eq_pfrc_cPBE_func, zeroin
#integer, external :: isOK_pset
#
#!open (3, FILE='logf.txt',FORM ='FORMATTED')
#    
#    do i=1,n
#        g_pset = psets(:,i)
#        if (isOK_pset(g_pset)<0) then
#            pfrc = -1.0
#            EXIT
#        endif
#
#        pf = zeroin(eq_pfrc_cPBE_func, 0.0, 1.0, tol, tol)
#
#        call eq_pfrc_cPBE(pf,g_pset,p3)
#        pfrc(1,i) = pf
#        pfrc(2,i) = p3(2)
#        pfrc(3,i) = p3(3)       
#        call get_pf_prior_cPBE(g_pset,prior)
#        pfrc(4,i) = prior
#    end do
#!close(3)
#
#End subroutine
#
#Function eq_pfrc_cPBE_func(pf)
#USE global_variables
#Implicit none
#real pf, temp(3)
#real eq_pfrc_cPBE_func
#    call eq_pfrc_cPBE(pf,g_pset,temp)
#    eq_pfrc_cPBE_func = temp(1)
#End Function
#
#subroutine eq_pfrc_cPBE(pf,pset,df)
#USE global_variables
#Implicit none
#real  pf,pset(12)
#real, INTENT(OUT):: df(3)
#real  pr,pc,pfn
#real pc2
#
#    call get_pr_cPBE(pf,pset,pr)
#    if (pr==0.0) then
#        if ( pset(mVA) > pset(mSA) ) then
#            pc = 1.0
#            call get_pf_prior_cPBE(pset,pfn)
#        elseif ( pset(mVA) < pset(mSA) ) then
#            pc = 0.0
#            pfn = 0.0
#        else
#            pc = 0.5
#            call get_pf_prior_cPBE(pset,pfn)
#            pfn = pfn / 2.0
#        endif
#    else    
#        call get_pc_cPBE(pr,pset,pc2)
#        call get_pf_numerator_cPBE(pr,pset,pfn)
#        pc = pc2 + pfn
#    endif
#
#!   df(1) = pf*pc - pfn
#    df(1) = pf - pfn/pc
#    df(2) = pr
#    df(3) = pc
#    
#End subroutine
#
#subroutine get_pr_cPBE(pf,pset,pr)
#USE global_variables
#USE Bivariate_Trivariate_Normal
#Implicit none
#real  pf,pset(12)
#real, INTENT(OUT):: pr
#real  Xmean, Xstd
#    if (pf==0.0) then
#        if ( pset(mCB) < pset(mVB) ) then
#            pr = 1.0
#        elseif ( pset(mCB) > pset(mVB) ) then
#            pr = 0.0
#        else
#            pr = 0.5
#        endif
#    else
#        Xmean = pf*pset(mWB) + (1-pf)* pset(mVB) - pset(mCB)
#        Xstd = (pf**2 + (1-pf)**2*abs(pset(sVVB)) + 2*pf*(1-pf)*pset(sVWB))**0.5
#        pr =  phid(Xmean / Xstd)
#    endif
#
#End subroutine
#
#subroutine get_pc_cPBE(pr,pset,pc2)
#USE global_variables
#USE Bivariate_Trivariate_Normal
#Implicit none
#real  pr,pset(12)
#real, INTENT(OUT):: pc2
#!real  Ymean, Ystd, Zmean, Zstd, YZcorr
#
#real  Umean, Ustd, Vmean, Vstd, UVcorr   !, pf_temp
#
#    Umean = pset(maA) - pset(mWA)
#    Ustd = ( 1 + pset(sWWA) - 2*pset(saWA) )**0.5
#    Vmean = (1-pr)*pset(mVA) + pr*pset(maA) - pset(mSA)
#    Vstd = ( (1-pr)**2*pset(sVVA) + pr**2 + 2*(1-pr)*pr*pset(sVaA) )**0.5
#    UVcorr = (1-pr)*(pset(sVaA) - pset(sVWA)) + pr*( 1 - pset(saWA) )
#    UVcorr = UVcorr / (Ustd*Vstd)
#    
#!   pf_temp = bvn(-Umean/Ustd,-Vmean/Vstd,UVcorr)
#!   if ((pc==0.0) .AND. (pf_temp>0)) then 
#!       pf = 1000   
#!   elseif ((pc==0.0) .AND. (pf_temp<0)) then 
#!       pf = -1000
#!   else
#        pc2 = bvn(-Umean/Ustd,-Vmean/Vstd,UVcorr)
#
#
#! original version
#!   Ymean = pset(mSA) - (1-pr)*pset(mVA) - pr*pset(maA)
#!   Ystd = ( (1-pr)**2*pset(sVVA) + pr**2 + 2*(1-pr)*pr*pset(sVaA) )**0.5
#!   Zmean = pset(mSA) -(1-pr)*pset(mVA) - pr*pset(mWA)
#!   Zstd = ( (1-pr)**2*pset(sVVA) + pr**2*pset(sWWA) + 2*(1-pr)*pr*pset(sVWA) )**0.5
#!   YZcorr = (1-pr)**2*pset(sVVA) +(1-pr)*pr*(pset(sVWA)+pset(sVaA))+pr**2*pset(saWA)
#!   YZcorr = YZcorr / (Ystd*Zstd)
#
#
#!   pc = 1- bvn(-Ymean/Ystd,-Zmean/Zstd,YZcorr)  ! note : -, -, +
#
#End subroutine
#
#subroutine get_pf_numerator_cPBE(pr,pset,pfn)
#USE global_variables
#USE Bivariate_Trivariate_Normal
#Implicit none
#real  pr,pset(12)
#real, INTENT(OUT):: pfn
#real  Umean, Ustd, Vmean, Vstd, UVcorr   !, pf_temp
#
#    Umean = pset(mWA) - pset(maA)
#    Ustd = ( pset(sWWA) + 1 - 2*pset(saWA) )**0.5
#    Vmean = (1-pr)*pset(mVA) + pr*pset(mWA) - pset(mSA)
#    Vstd = ( (1-pr)**2*pset(sVVA) + pr**2*pset(sWWA) + 2*(1-pr)*pr*pset(sVWA) )**0.5
#    UVcorr = (1-pr)*(pset(sVWA) - pset(sVaA)) + pr*( pset(sWWA) - pset(saWA) )
#    UVcorr = UVcorr / (Ustd*Vstd)
#    
#!   pf_temp = bvn(-Umean/Ustd,-Vmean/Vstd,UVcorr)
#!   if ((pc==0.0) .AND. (pf_temp>0)) then 
#!       pf = 1000   
#!   elseif ((pc==0.0) .AND. (pf_temp<0)) then 
#!       pf = -1000
#!   else
#        pfn = bvn(-Umean/Ustd,-Vmean/Vstd,UVcorr)
#!       pf = pf_temp
#!   end if
#    
#End subroutine
#
#subroutine get_pf_prior_cPBE(pset,prior)
#USE global_variables
#USE Bivariate_Trivariate_Normal
#Implicit none
#real  pset(12)
#real, INTENT(OUT):: prior
#real  Umean, Ustd
#
#    Umean = pset(mWA) - pset(maA)
#    Ustd = ( pset(sWWA) + 1 - 2*pset(saWA) )**0.5
#    prior = phid( Umean / Ustd )
#End subroutine
#
#subroutine phid_test(z,p)   ! normal cdf
#USE Bivariate_Trivariate_Normal
#Implicit none
#real  z
#real, INTENT(OUT):: p
#    p = phid(z)
#    
#end subroutine
#
#Function isOK_pset(pset)
#USE global_variables
#USE eigen
#Implicit none
#real pset(14)
#integer isOK_pset, flag 
#real V(3,3), eigenv(3), trace, sume
#integer ier
#
#    flag=1
#    if ((pset(sVVA)<0) .OR. (pset(sWWA)<0) .OR. (pset(sVVB)<0)) then
#        flag=-1
#    else
#        if (pset(sVVA)==0) then
#            if ( (pset(sVaA)/=0.0) .OR. (pset(sVWA)/=0.0) ) flag=-2
#        else
#            if (abs(pset(sVaA)/pset(sVVA))**0.5 > 1) flag=-3
#            if (abs(pset(sVWA)/(pset(sVVA)*pset(sWWA))**0.5) > 1.0) flag=-4
#        endif
#        
#        if (abs(pset(saWA)/(pset(sWWA))**0.5) > 1) flag=-5
#        if (abs(pset(sVWB)/(pset(sVVB))**0.5) > 1) flag=-6
#        
#    endif
#
#    V = reshape( source = (/ pset(sVVA), pset(sVaA), pset(sVWA), &
#                            pset(sVaA),        1.0, pset(saWA), &
#                            pset(sVWA), pset(saWA), pset(sWWA) /),&
#                 shape = (/3,3/))
#    call kaiser(V,3, 3, eigenv, trace, sume, ier)
#!write (3,*), eigenv
#
#    if (any(eigenv<0)) flag = -7     
#    isOK_pset = flag
#End Function
#
#subroutine init_cPBE_dll()
#USE global_variables
#Implicit none
#
#    mSA=1
#    mVA=2
#    maA=3
#    mWA=4
#    sVVA=5
#    sVaA=6
#    sVWA=7
#    saWA=8
#    sWWA=9
#    mCB=10
#    mVB=11
#    mWB=12
#    sVVB=13
#    sVWB=14
#end subroutine
